"""
This code is from: https://github.com/tomaarsen/TTSTextNormalization/blob/master/converters/Digit.py

"""

import re


# Function to convert numbers to words, adapted from GPT generation
def number_to_words(num):

    # Create a dictionary for mapping digits to words
    digits = {
        0: 'zero', 1: 'one', 2: 'two', 3: 'three', 4: 'four',
        5: 'five', 6: 'six', 7: 'seven', 8: 'eight', 9: 'nine'
    }

    # Create a dictionary for mapping two-digit numbers to words
    two_digits = {
        10: 'ten', 11: 'eleven', 12: 'twelve', 13: 'thirteen', 14: 'fourteen',
        15: 'fifteen', 16: 'sixteen', 17: 'seventeen', 18: 'eighteen', 19: 'nineteen',
        20: 'twenty', 30: 'thirty', 40: 'forty', 50: 'fifty', 60: 'sixty',
        70: 'seventy', 80: 'eighty', 90: 'ninety'
    }

    # Create a dictionary for mapping powers of 10 to words
    powers_of_10 = {
        100: 'hundred', 1000: 'thousand', 1000000: 'million',
        1000000000: 'billion', 1000000000000: 'trillion'
    }

    # Handle negative numbers
    if num < 0:
        return "minus " + number_to_words(abs(num))

    # Handle numbers from 0 to 9
    if num < 10:
        return digits[num]

    # Handle numbers from 10 to 99
    if num < 100:
        if num in two_digits:
            return two_digits[num]
        else:
            return two_digits[num // 10 * 10] + " " + digits[num % 10]

    # Handle numbers from 100 to 999
    if num < 1000:
        hundreds = num // 100
        remainder = num % 100
        if remainder == 0:
            return digits[hundreds] + " " + powers_of_10[100]
        else:
            return digits[hundreds] + " " + powers_of_10[100] + " and " + number_to_words(remainder)

    # Handle numbers greater than or equal to 1000
    for power in sorted(powers_of_10.keys(), reverse=True):
        if num >= power:
            quotient = num // power
            remainder = num % power
            if remainder == 0:
                return number_to_words(quotient) + " " + powers_of_10[power]
            else:
                return number_to_words(quotient) + " " + powers_of_10[power] + " " + number_to_words(remainder)

    return "infinity"

class ENGNormalizer:

    def __init__(self):
        super().__init__()
        # Regex used to filter out non digits
        self.filter_regex = re.compile("[^0-9]")
        # Translation dict to convert digits to text
        self.trans_dict = {
            "0": "zero",
            "1": "one",
            "2": "two",
            "3": "three",
            "4": "four",
            "5": "five",
            "6": "six",
            "7": "seven",
            "8": "eight",
            "9": "nine"
        }

        self.punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'


    def __call__(self, text: str) -> str:

        text = text.translate(str.maketrans('', '', self.punctuation)).lower()

        tokens = text.split()

        res = []
        for token in tokens:
            if str.isdigit(token):
                # 1 Filter out anything that isn't a digit
                token = self.filter_regex.sub("", text)
                # 2 Check for special case
                if token == "007":
                    return "double o seven"

                token = number_to_words(int(token))

            res.append(token)

        return " ".join(res)



from transphone.model.utils import pad_list
import torch
from torch.utils.data import DataLoader


def collate(xy_lst):
    x_lst = [xy[0] for xy in xy_lst]
    y_lst = [xy[1] for xy in xy_lst]

    x = pad_list(x_lst)
    y = pad_list(y_lst)

    return x,y

def read_loader(dataset, batch_size=32):

    loader = DataLoader(dataset, shuffle=True, batch_size=batch_size, collate_fn=collate)
    return loader


"""Basic Epitran class for G2P in most languages."""
import logging
import os.path
import sys
import csv
import unicodedata
from collections import defaultdict
from typing import DefaultDict, Callable  # pylint: disable=unused-import
from phonepiece.epitran import read_epitran_g2p
from phonepiece.lexicon import read_lexicon
from phonepiece.inventory import read_inventory
import re
from transphone.lang.base_tokenizer import BaseTokenizer
from epitran.ppprocessor import PrePostProcessor
from phonepiece.ipa import read_ipa
import epitran

logger = logging.getLogger('epitran')

customized_epitran_dict = {
    'spa': 'spa-Latn',
    'deu': 'deu-Latn',
    'ita': 'ita-Latn',
    'rus': 'rus-Cyrl',
    'fra': 'fra-Latn',
}

raw_epitran_dict = {
     'tur': 'tur-Latn',
     'vie': 'vie-Latn',
     'aar': 'aar-Latn',
     'got': 'got-Latn',
     'lsm': 'lsm-Latn',
     'swa': 'swa-Latn',
     'aii': 'aii-Syrc',
     'hak': 'hak-Latn',
     'ltc': 'ltc-Latn-bax',
     'swe': 'swe-Latn',
     'amh': 'amh-Ethi',
     'hat': 'hat-Latn-bab',
     'tam': 'tam-Taml',
     'hau': 'hau-Latn',
     'mal': 'mal-Mlym',
     'hin': 'hin-Deva',
     'mar': 'mar-Deva',
     'tel': 'tel-Telu',
     'ara': 'ara-Arab',
     'hmn': 'hmn-Latn',
     'mlt': 'mlt-Latn',
     'tgk': 'tgk-Cyrl',
     'ava': 'ava-Cyrl',
     'hrv': 'hrv-Latn',
     'mon': 'mon-Cyrl-bab',
     'tgl': 'tgl-Latn',
     'aze': 'aze-Latn',
     'hsn': 'hsn-Latn',
     'mri': 'mri-Latn',
     'hun': 'hun-Latn',
     'msa': 'msa-Latn',
     'tha': 'tha-Thai',
     'ilo': 'ilo-Latn',
     'mya': 'mya-Mymr',
     'tir': 'tir-Ethi',
     'ind': 'ind-Latn',
     'nan': 'nan-Latn',
     'ita': 'ita-Latn',
     'bxk': 'bxk-Latn',
     'jam': 'jam-Latn',
     'nld': 'nld-Latn',
     'tpi': 'tpi-Latn',
     'cat': 'cat-Latn',
     'jav': 'jav-Latn',
     'nya': 'nya-Latn',
     'tuk': 'tuk-Latn',
     'ceb': 'ceb-Latn',
     'jpn': 'jpn-Ktkn-red',
     'ood': 'ood-Latn-sax',
     'ces': 'ces-Latn',
     'cjy': 'cjy-Latn',
     'ori': 'ori-Orya',
     'ckb': 'ckb-Arab',
     'orm': 'orm-Latn',
     'cmn': 'cmn-Latn',
     'kat': 'kat-Geor',
     'pan': 'pan-Guru',
     'uew': 'uew',
     'kaz': 'kaz-Latn',
     'pii': 'pii-latn_Wiktionary',
     'uig': 'uig-Arab',
     'csb': 'csb-Latn',
     'ukr': 'ukr-Cyrl',
     'deu': 'deu-Latn',
     'pol': 'pol-Latn',
     'urd': 'urd-Arab',
     'kbd': 'kbd-Cyrl',
     'por': 'por-Latn',
     'uzb': 'uzb-Latn',
     'khm': 'khm-Khmr',
     'ron': 'ron-Latn',
     'fas': 'fas-Arab',
     'kin': 'kin-Latn',
     'run': 'run-Latn',
     'kir': 'kir-Latn',
     'rus': 'rus-Cyrl',
     'wuu': 'wuu-Latn',
     'sag': 'sag-Latn',
     'xho': 'xho-Latn',
     'sin': 'sin-Sinh',
     'yor': 'yor-Latn',
     'kmr': 'kmr-Latn',
     'sna': 'sna-Latn',
     'yue': 'yue-Latn',
     'som': 'som-Latn',
     'zha': 'zha-Latn',
     'ful': 'ful-Latn',
     'lao': 'lao-Laoo-prereform',
     'spa': 'spa-Latn-eu',
     'zul': 'zul-Latn',
     'gan': 'gan-Latn',
     'sqi': 'sqi-Latn',
     'lij': 'lij-Latn'
}


def read_epitran_tokenizer(lang_id, g2p_model=None, device=None, use_lexicon=True):
    if lang_id in customized_epitran_dict:
        return read_customized_epitran_tokenizer(customized_epitran_dict[lang_id], use_lexicon)
    elif lang_id in raw_epitran_dict:
        return read_raw_epitran_tokenizer(raw_epitran_dict[lang_id], use_lexicon)
    else:
        raise ValueError('Unknown epitran id: {}'.format(lang_id_or_epi_id))


def read_raw_epitran_tokenizer(lang_id_or_epi_id, use_lexicon=True):

    if '-' in lang_id_or_epi_id:
        lang_id, writing_system = lang_id_or_epi_id.split('-', 1)
    else:
        lang_id = lang_id_or_epi_id
        writing_system = 'Latn'

    if use_lexicon:
        lexicon = read_lexicon(lang_id)
    else:
        lexicon = {}

    return RawEpitranTokenizer(lang_id, writing_system, lexicon)


def read_customized_epitran_tokenizer(lang_id_or_epi_id, use_lexicon=True):

    if '-' in lang_id_or_epi_id:
        lang_id, writing_system = lang_id_or_epi_id.split('-', 1)
    else:
        lang_id = lang_id_or_epi_id
        writing_system = 'Latn'

    if use_lexicon:
        lexicon = read_lexicon(lang_id)
    else:
        lexicon = {}

    return CustomizedEpitranTokenizer(lang_id, writing_system, lexicon)


class CustomizedEpitranTokenizer(BaseTokenizer):

    def __init__(self, lang_id, writing_system=None, lexicon=None):
        super().__init__(lang_id, None)

        #self.lexicon = read_lexicon(lang_id)

        """Constructor"""
        self.lang_id = lang_id
        self.writing_system = writing_system

        if writing_system:
            lang_id = lang_id + '-' + writing_system

        if not lexicon:
            lexicon = {}

        self.lexicon = lexicon

        self.g2p = read_epitran_g2p(lang_id)
        self.inv = read_inventory(self.lang_id)
        self.ipa = read_ipa()

        self.regexp = self._construct_regex(self.g2p.keys())
        self.nils = defaultdict(int)

        self.cache = defaultdict(list)

        self.preprocessor = PrePostProcessor(lang_id, 'pre', False)
        #self.postprocessor = PrePostProcessor(lang_id, 'post', False)

    def _construct_regex(self, g2p_keys):
        """Build a regular expression that will greadily match segments from
           the mapping table.
        """
        graphemes = sorted(g2p_keys, key=len, reverse=True)
        return re.compile(f"({r'|'.join(graphemes)})", re.I)

    def match_word(self, text, verbose=False):
        ipa_lst = []
        while text:
            logger.debug('text=', repr(list(text)))
            if verbose:
                print('text=', repr(list(text)))

            m = self.regexp.match(text)
            if m:
                source = m.group(0)
                try:
                    targets = self.g2p[source]
                    if verbose:
                        print(source, ' -> ', targets)
                except KeyError:
                    logger.debug("source = '%s''", source)
                    logger.debug("self.g2p[source] = %s'", self.g2p[source])
                    targets = []
                except IndexError:
                    logger.debug("self.g2p[source]= %s", self.g2p[source])
                    targets = []

                ipa_lst.extend(targets)
                text = text[len(source):]
            else:
                self.nils[text[0]] += 2
                text = text[1:]
        ipa_lst = self.inv.remap(ipa_lst)
        return ipa_lst


    def tokenize(self, text: str, verbose: bool=False):
        text = text.lower()

        ipa_lst = []

        for word in text.split():

            if word in self.cache:
                ipa_lst.extend(self.cache[word])

            elif word in self.lexicon:
                phonemes = self.lexicon[word]
                ipa_lst.extend(phonemes)
                self.cache[word] = phonemes
                log = f"lexicon {word} -> {phonemes}"
                self.logger.info(log)
                if verbose:
                    print(log)
            else:
                norm_word = unicodedata.normalize('NFC', word)

                norm_word = self.preprocessor.process(norm_word)

                word_ipa_lst = self.match_word(norm_word, verbose)

                log = f"rule raw: {word} -> norm: {norm_word} -> {word_ipa_lst}"
                self.logger.info(log)
                if verbose:
                    print(log)

                #word_ipa_lst = self.ipa.tokenize(self.postprocessor.process(''.join(word_ipa_lst)))

                self.cache[word] = word_ipa_lst
                ipa_lst.extend(self.cache[word])

        return ipa_lst


class RawEpitranTokenizer(BaseTokenizer):

    def __init__(self, lang_id, writing_system='Latin', lexicon=None):
        super().__init__(lang_id, None)

        #self.lexicon = read_lexicon(lang_id)

        """Constructor"""
        self.lang_id = lang_id
        self.writing_system = writing_system

        if writing_system:
            lang_id = lang_id + '-' + writing_system

        if not lexicon:
            lexicon = {}

        self.lexicon = lexicon

        self.g2p = epitran.Epitran(lang_id)
        self.inv = read_inventory(self.lang_id)
        self.ipa = read_ipa()

        self.cache = defaultdict(list)


    def tokenize(self, text: str, verbose: bool=False):
        text = text.lower()

        ipa_lst = []

        for word in text.split():

            if word in self.cache:
                ipa_lst.extend(self.cache[word])

            elif word in self.lexicon:
                phonemes = self.lexicon[word]
                ipa_lst.extend(phonemes)
                self.cache[word] = phonemes
                log = f"lexicon {word} -> {phonemes}"
                self.logger.info(log)
                if verbose:
                    print(log)
            else:

                word_ipa_lst = self.g2p.trans_list(word)
                word_ipa_lst = self.inv.remap(word_ipa_lst, broad=True)

                log = f"rule raw: {word} -> remap {word_ipa_lst}"
                self.logger.info(log)
                if verbose:
                    print(log)

                #word_ipa_lst = self.ipa.tokenize(self.postprocessor.process(''.join(word_ipa_lst)))

                self.cache[word] = word_ipa_lst
                ipa_lst.extend(self.cache[word])

        return ipa_lst


from transphone.bin.download_model import download_model
from transphone.model.utils import resolve_model_name
from transphone.tokenizer import read_tokenizer
from pathlib import Path
import argparse
import tqdm


if __name__ == '__main__':

    parser = argparse.ArgumentParser('running transphone g2p model')
    parser.add_argument('-m', '--model', type=str, default='latest',
                        help='specify which model to use. default is to use the latest local model')
    parser.add_argument('-l', '--lang', type=str, default='eng',
                        help='specify which language inventory to use for recognition. default is to use all phone inventory')
    parser.add_argument('-i', '--input', type=str, required=True, help='specify your input wav file/directory')
    parser.add_argument('-o', '--output', type=str, default='stdout',
                        help='specify output file. the default will be stdout')
    parser.add_argument('-f', '--format', type=str, default='text', help='kaldi or text')
    parser.add_argument('-c', '--combine', type=bool, default=False,
                        help='write outputs by including both grapheme inputs and phonemes in the same line, delimited by space')

    args = parser.parse_args()

    # resolve model's name
    model_name = resolve_model_name(args.model)

    # format
    file_format = args.format

    if args.combine:
        assert file_format == 'text'

    # download specified model automatically if no model exists
    download_model(model_name)

    # create model
    tokenizer = read_tokenizer(args.lang, g2p_model=model_name)

    # output file descriptor
    output_fd = None
    if args.output != 'stdout':
        output_fd = open(args.output, 'w', encoding='utf-8')

    # input file/path
    input_path = Path(args.input)

    for line in tqdm.tqdm(open(input_path, 'r').readlines(), disable=args.output=='stdout'):
        fields = line.strip().split()

        utt_id = None

        if file_format == 'text':
            text = ' '.join(fields)
        else:
            text = ' '.join(fields[1:])
            utt_id = fields[0]

        phonemes = tokenizer.tokenize(text)
        line_output = ' '.join(phonemes)

        if args.combine:
            line_output = text + '\t' + line_output

        if utt_id is not None:
            line_output = utt_id + ' ' + line_output

        if output_fd:
            output_fd.write(line_output + '\n')
        else:
            print(line_output)

    if output_fd:
        output_fd.close()


import unittest
from transphone.tokenizer import read_tokenizer


class TestTokenizer(unittest.TestCase):

    def test_eng_tokenizer(self):
        eng = read_tokenizer('eng')

        self.assertEqual(eng.tokenize_words('hello world'), ['hello', 'world'])
        self.assertEqual(eng.tokenize_words('hello world!!  '), ['hello', 'world'])
        self.assertEqual(eng.tokenize('hello world'), ['h', 'ʌ', 'l', 'o', 'w', 'w', 'ɹ̩', 'l', 'd'])

    def test_jpn_tokenizer(self):
        jpn = read_tokenizer('jpn')
        self.assertEqual(jpn.tokenize_words('こんにちは世界'), ['こんにちは', '世界'])
        self.assertEqual(jpn.tokenize('こんにちは世界'), ['k', 'o', 'N', 'n', 'i', 'ch', 'i', 'w', 'a', 's', 'e', 'k', 'a', 'i'])
        self.assertEqual(jpn.tokenize('2023年'), ['n', 'i', 's', 'e', 'N', 'n', 'i', 'j', 'u', 'u', 's', 'a', 'N', 'n', 'e', 'N'])
        self.assertEqual(jpn.tokenize('ＵTＦとABC。'),  ['y', 'uː', 't', 'iː', 'e', 'f', 'u', 't', 'o', 'eː', 'b', 'iː', 'sh', 'iː'])

    def test_spa_tokenizer(self):

        spa = read_tokenizer('spa')
        self.assertEqual(spa.tokenize('hola hola'), ['o', 'l', 'a', 'o', 'l', 'a'])
        self.assertEqual(spa.tokenize('español'),  ['e', 's', 'p', 'a', 'ɲ', 'o', 'l'])

        spa = read_tokenizer('spa', use_lexicon=False)
        self.assertEqual(spa.tokenize('hola hola'), ['o', 'l', 'a', 'o', 'l', 'a'])
        self.assertEqual(spa.tokenize('español'),  ['e', 's', 'p', 'a', 'ɲ', 'o', 'l'])

    def test_fra_tokenizer(self):

        fra = read_tokenizer('fra')
        self.assertEqual(fra.tokenize('français'), ['f', 'ʁ', 'ɑ̃', 's', 'ɛ'])

        fra = read_tokenizer('fra', use_lexicon=False)
        self.assertEqual(fra.tokenize('français'), ['f', 'ʁ', 'ɑ̃', 's', 'ɛ'])

    def test_deu_tokenizer(self):

        deu = read_tokenizer('deu')
        self.assertEqual(deu.tokenize('Deutsche'), ['d', 'o', 'i', 't͡ʃ', 'ə'])

        deu = read_tokenizer('deu', use_lexicon=False)
        self.assertEqual(deu.tokenize('Deutsche'),['d', 'o', 'i', 't͡ʃ', 'ə'])

    def test_ita_tokenizer(self):
        ita = read_tokenizer('ita')
        self.assertEqual(ita.tokenize('Italia'), ['i', 't', 'a', 'l', 'j', 'a'])

        # g2p is slightly different here
        ita = read_tokenizer('ita', use_lexicon=False)
        self.assertEqual(ita.tokenize('Italia'), ['i', 't', 'a', 'l', 'i', 'a'])

    def test_tur_tokenizer(self):

        tur = read_tokenizer('tur', use_lexicon=False)
        self.assertEqual(tur.tokenize('Türkçe'), ['t', 'y', 'ɾ', 'k', 't͡ʃ', 'e'])

        tur = read_tokenizer('tur')
        self.assertEqual(tur.tokenize('Türkçe'), ['t', 'y', 'ɾ', 'k', 't͡ʃ', 'e'])



from phonepiece.lang import normalize_lang_id
from phonepiece.lexicon import read_lexicon
from transphone.lang.base_tokenizer import BaseTokenizer


def read_g2p_tokenizer(lang_id, g2p_model='latest', device=None):
    lang_id = normalize_lang_id(lang_id)
    return G2PTokenizer(lang_id, g2p_model, device)

class G2PTokenizer(BaseTokenizer):

    def __init__(self, lang_id, g2p_model='latest', device=None):
        super().__init__(lang_id, g2p_model, device)

        self.lexicon = read_lexicon(lang_id)


    def tokenize(self, text, use_g2p=True, use_space=False, verbose=False):

        norm_text = text.translate(str.maketrans('', '', self.punctuation)).lower()
        log = f"normalization: {text} -> {norm_text}"
        self.logger.info(log)

        if verbose:
            print(log)

        text = norm_text

        result = []

        for word in text.split():
            if word in self.cache:
                result.extend(self.cache[word])
            elif word in self.lexicon:
                phonemes = self.lexicon[word]
                result.extend(phonemes)
                self.cache[word] = phonemes
                log = f"lexicon {word} -> {phonemes}"
                self.logger.info(log)
                if verbose:
                    print(log)
            else:
                phonemes = self.g2p.inference_batch(word, self.lang_id, verbose=verbose)
                remapped_phonemes = self.inventory.remap(phonemes)

                log = f"g2p batch mode: {word} ->  {remapped_phonemes}"
                self.logger.info(log)
                if verbose:
                    print(log)
                self.cache[word] = remapped_phonemes
                result.extend(remapped_phonemes)
            if use_space:
                result.append('<SPACE>')

        return result



# -*- coding: utf-8 -*-
import unicodedata
from .conv_table import (H2K_TABLE, H2HK_TABLE, K2H_TABLE, H2Z_A, H2Z_AD,
                         H2Z_AK, H2Z_D, H2Z_K, H2Z_DK, H2Z_ALL, Z2H_A, Z2H_AD,
                         Z2H_AK, Z2H_D, Z2H_K, Z2H_DK, Z2H_ALL, KANA2HEP, HEP2KANA)

consonants = frozenset('sdfghjklqwrtypzxcvbnm')

def _exclude_ignorechar(ignore, conv_map):
    for character in map(ord, ignore):
        conv_map[character] = character
    return conv_map


def _convert(text, conv_map):
    return text.translate(conv_map)


def hira2kata(text, ignore=''):
    """Convert Hiragana to Full-width (Zenkaku) Katakana.
    Parameters
    ----------
    text : str
        Hiragana string.
    ignore : str
        Characters to be ignored in converting.
    Return
    ------
    str
        Katakana string.
    Examples
    --------
    >>> print(jaconv.hira2kata('ともえまみ'))
    トモエマミ
    >>> print(jaconv.hira2kata('まどまぎ', ignore='ど'))
    マどマギ
    """
    if ignore:
        h2k_map = _exclude_ignorechar(ignore, H2K_TABLE.copy())
        return _convert(text, h2k_map)
    return _convert(text, H2K_TABLE)


def hira2hkata(text, ignore=''):
    """Convert Hiragana to Half-width (Hankaku) Katakana
    Parameters
    ----------
    text : str
        Hiragana string.
    ignore : str
        Characters to be ignored in converting.
    Return
    ------
    str
        Half-width Katakana string.
    Examples
    --------
    >>> print(jaconv.hira2hkata('ともえまみ'))
    ﾄﾓｴﾏﾐ
    >>> print(jaconv.hira2hkata('ともえまみ', ignore='み'))
    ﾄﾓｴﾏみ
    """
    if ignore:
        h2hk_map = _exclude_ignorechar(ignore, H2HK_TABLE.copy())
        return _convert(text, h2hk_map)
    return _convert(text, H2HK_TABLE)


def kata2hira(text, ignore=''):
    """Convert Full-width Katakana to Hiragana
    Parameters
    ----------
    text : str
        Full-width Katakana string.
    ignore : str
        Characters to be ignored in converting.
    Return
    ------
    str
        Hiragana string.
    Examples
    --------
    >>> print(jaconv.kata2hira('巴マミ'))
    巴まみ
    >>> print(jaconv.kata2hira('マミサン', ignore='ン'))
    まみさン
    """
    if ignore:
        k2h_map = _exclude_ignorechar(ignore, K2H_TABLE.copy())
        return _convert(text, k2h_map)
    return _convert(text, K2H_TABLE)


def h2z(text, ignore='', kana=True, ascii=False, digit=False):
    """Convert Half-width (Hankaku) Katakana to Full-width (Zenkaku) Katakana
    Parameters
    ----------
    text : str
        Half-width Katakana string.
    ignore : str
        Characters to be ignored in converting.
    kana : bool
        Either converting Kana or not.
    ascii : bool
        Either converting ascii or not.
    digit : bool
        Either converting digit or not.
    Return
    ------
    str
        Full-width Katakana string.
    Examples
    --------
    >>> print(jaconv.h2z('ﾃｨﾛﾌｨﾅｰﾚ'))
    ティロフィナーレ
    >>> print(jaconv.h2z('ﾃｨﾛﾌｨﾅｰﾚ', ignore='ｨ'))
    テｨロフｨナーレ
    >>> print(jaconv.h2z('abcd', ascii=True))
    ＡＢＣＤ
    >>> print(jaconv.h2z('1234', digit=True))
    １２３４
    """
    def _conv_dakuten(text):
        """Convert Hankaku Dakuten Kana to Zenkaku Dakuten Kana
        """
        text = text.replace("ｶﾞ", "ガ").replace("ｷﾞ", "ギ")
        text = text.replace("ｸﾞ", "グ").replace("ｹﾞ", "ゲ")
        text = text.replace("ｺﾞ", "ゴ").replace("ｻﾞ", "ザ")
        text = text.replace("ｼﾞ", "ジ").replace("ｽﾞ", "ズ")
        text = text.replace("ｾﾞ", "ゼ").replace("ｿﾞ", "ゾ")
        text = text.replace("ﾀﾞ", "ダ").replace("ﾁﾞ", "ヂ")
        text = text.replace("ﾂﾞ", "ヅ").replace("ﾃﾞ", "デ")
        text = text.replace("ﾄﾞ", "ド").replace("ﾊﾞ", "バ")
        text = text.replace("ﾋﾞ", "ビ").replace("ﾌﾞ", "ブ")
        text = text.replace("ﾍﾞ", "ベ").replace("ﾎﾞ", "ボ")
        text = text.replace("ﾊﾟ", "パ").replace("ﾋﾟ", "ピ")
        text = text.replace("ﾌﾟ", "プ").replace("ﾍﾟ", "ペ")
        return text.replace("ﾎﾟ", "ポ").replace("ｳﾞ", "ヴ")

    if ascii:
        if digit:
            if kana:
                h2z_map = H2Z_ALL
            else:
                h2z_map = H2Z_AD
        elif kana:
            h2z_map = H2Z_AK
        else:
            h2z_map = H2Z_A
    elif digit:
        if kana:
            h2z_map = H2Z_DK
        else:
            h2z_map = H2Z_D
    else:
        h2z_map = H2Z_K
    if kana:
        text = _conv_dakuten(text)
    if ignore:
        h2z_map = _exclude_ignorechar(ignore, h2z_map.copy())
    return _convert(text, h2z_map)


def z2h(text, ignore='', kana=True, ascii=False, digit=False):
    """Convert Full-width (Zenkaku) Katakana to Half-width (Hankaku) Katakana
    Parameters
    ----------
    text : str
        Full-width Katakana string.
    ignore : str
        Characters to be ignored in converting.
    kana : bool
        Either converting Kana or not.
    ascii : bool
        Either converting ascii or not.
    digit : bool
        Either converting digit or not.
    Return
    ------
    str
        Half-width Katakana string.
    Examples
    --------
    >>> print(jaconv.z2h('ティロフィナーレ'))
    ﾃｨﾛﾌｨﾅｰﾚ
    >>> print(jaconv.z2h('ティロフィナーレ', ignore='ィ'))
    ﾃィﾛﾌィﾅｰﾚ
    >>> print(jaconv.z2h('ＡＢＣＤ', ascii=True))
    abcd
    >>> print(jaconv.z2h('１２３４', digit=True))
    1234
    """
    if ascii:
        if digit:
            if kana:
                z2h_map = Z2H_ALL
            else:
                z2h_map = Z2H_AD
        elif kana:
            z2h_map = Z2H_AK
        else:
            z2h_map = Z2H_A
    elif digit:
        if kana:
            z2h_map = Z2H_DK
        else:
            z2h_map = Z2H_D
    else:
        z2h_map = Z2H_K
    if ignore:
        z2h_map = _exclude_ignorechar(ignore, z2h_map.copy())
    return _convert(text, z2h_map)


def normalize(text, mode='NFKC'):
    """Convert Half-width (Hankaku) Katakana to Full-width (Zenkaku) Katakana,
    Full-width (Zenkaku) ASCII and DIGIT to Half-width (Hankaku) ASCII
    and DIGIT.
    Additionally, Full-width wave dash (〜) etc. are normalized
    Parameters
    ----------
    text : str
        Source string.
    mode : str
        Unicode normalization mode.
    Return
    ------
    str
        Normalized string.
    Examples
    --------
    >>> print(jaconv.normalize('ﾃｨﾛ･フィナ〜レ', 'NFKC'))
    ティロ・フィナーレ
    """
    text = text.replace('〜', 'ー').replace('～', 'ー')
    text = text.replace("’", "'").replace('”', '"').replace('“', '``')
    text = text.replace('―', '-').replace('‐', '-').replace('˗', '-').replace('֊', '-')
    text = text.replace('‐', '-').replace('‑', '-').replace('‒', '-').replace('–', '-')
    text = text.replace('⁃', '-').replace('⁻', '-').replace('₋', '-').replace('−', '-')
    text = text.replace('﹣', 'ー').replace('－', 'ー').replace('—', 'ー').replace('―', 'ー')
    text = text.replace('━', 'ー').replace('─', 'ー')
    return unicodedata.normalize(mode, text)


def kana2alphabet(text):
    """Convert Hiragana to hepburn-style alphabets
    Parameters
    ----------
    text : str
        Hiragana string.
    Return
    ------
    str
        Hepburn-style alphabets string.
    Examples
    --------
    >>> print(jaconv.kana2alphabet('まみさん'))
    mamisan
    """
    text = text.replace('きゃ', 'kya').replace('きゅ', 'kyu').replace('きょ', 'kyo')
    text = text.replace('ぎゃ', 'gya').replace('ぎゅ', 'gyu').replace('ぎょ', 'gyo')
    text = text.replace('しゃ', 'sha').replace('しゅ', 'shu').replace('しょ', 'sho')
    text = text.replace('じゃ', 'ja').replace('じゅ', 'ju').replace('じょ', 'jo')
    text = text.replace('ちゃ', 'cha').replace('ちゅ', 'chu').replace('ちょ', 'cho')
    text = text.replace('にゃ', 'nya').replace('にゅ', 'nyu').replace('にょ', 'nyo')
    text = text.replace('ふぁ', 'fa').replace('ふぃ', 'fi').replace('ふぇ', 'fe')
    text = text.replace('ふぉ', 'fo')
    text = text.replace('ひゃ', 'hya').replace('ひゅ', 'hyu').replace('ひょ', 'hyo')
    text = text.replace('みゃ', 'mya').replace('みゅ', 'myu').replace('みょ', 'myo')
    text = text.replace('りゃ', 'rya').replace('りゅ', 'ryu').replace('りょ', 'ryo')
    text = text.replace('びゃ', 'bya').replace('びゅ', 'byu').replace('びょ', 'byo')
    text = text.replace('ぴゃ', 'pya').replace('ぴゅ', 'pyu').replace('ぴょ', 'pyo')
    text = text.replace('が', 'ga').replace('ぎ', 'gi').replace('ぐ', 'gu')
    text = text.replace('げ', 'ge').replace('ご', 'go').replace('ざ', 'za')
    text = text.replace('じ', 'ji').replace('ず', 'zu').replace('ぜ', 'ze')
    text = text.replace('ぞ', 'zo').replace('だ', 'da').replace('ぢ', 'ji')
    text = text.replace('づ', 'zu').replace('で', 'de').replace('ど', 'do')
    text = text.replace('ば', 'ba').replace('び', 'bi').replace('ぶ', 'bu')
    text = text.replace('べ', 'be').replace('ぼ', 'bo').replace('ぱ', 'pa')
    text = text.replace('ぴ', 'pi').replace('ぷ', 'pu').replace('ぺ', 'pe')
    text = text.replace('ぽ', 'po')
    text = text.replace('か', 'ka').replace('き', 'ki').replace('く', 'ku')
    text = text.replace('け', 'ke').replace('こ', 'ko').replace('さ', 'sa')
    text = text.replace('し', 'shi').replace('す', 'su').replace('せ', 'se')
    text = text.replace('そ', 'so').replace('た', 'ta').replace('ち', 'chi')
    text = text.replace('つ', 'tsu').replace('て', 'te').replace('と', 'to')
    text = text.replace('な', 'na').replace('に', 'ni').replace('ぬ', 'nu')
    text = text.replace('ね', 'ne').replace('の', 'no').replace('は', 'ha')
    text = text.replace('ひ', 'hi').replace('ふ', 'fu').replace('へ', 'he')
    text = text.replace('ほ', 'ho').replace('ま', 'ma').replace('み', 'mi')
    text = text.replace('む', 'mu').replace('め', 'me').replace('も', 'mo')
    text = text.replace('ら', 'ra').replace('り', 'ri').replace('る', 'ru')
    text = text.replace('れ', 're').replace('ろ', 'ro')
    text = text.replace('や', 'ya').replace('ゆ', 'yu').replace('よ', 'yo')
    text = text.replace('わ', 'wa').replace('ゐ', 'wi').replace('を', 'wo')
    text = text.replace('ゑ', 'we')
    text = _convert(text, KANA2HEP)
    while 'っ' in text:
        text = list(text)
        tsu_pos = text.index('っ')
        if len(text) <= tsu_pos + 1:
            return ''.join(text[:-1]) + 'xtsu'
        if tsu_pos == 0:
            text[tsu_pos] = 'xtsu'
        else:
            text[tsu_pos] = text[tsu_pos + 1]
        text = ''.join(text)
    return text


def alphabet2kana(text):
    """Convert alphabets to Hiragana
    Parameters
    ----------
    text : str
        Alphabets string.
    Return
    ------
    str
        Hiragana string.
    Examples
    --------
    >>> print(jaconv.alphabet2kana('mamisan'))
    まみさん
    """
    text = text.replace('kya', 'きゃ').replace('kyu', 'きゅ').replace('kyo', 'きょ')
    text = text.replace('gya', 'ぎゃ').replace('gyu', 'ぎゅ').replace('gyo', 'ぎょ')
    text = text.replace('sha', 'しゃ').replace('shu', 'しゅ').replace('sho', 'しょ')
    text = text.replace('zya', 'じゃ').replace('zyu', 'じゅ').replace('zyo', 'じょ')
    text = text.replace('zyi', 'じぃ').replace('zye', 'じぇ')
    text = text.replace('ja', 'じゃ').replace('ju', 'じゅ').replace('jo', 'じょ')
    text = text.replace('jya', 'じゃ').replace('jyu', 'じゅ').replace('jyo', 'じょ')
    text = text.replace('cha', 'ちゃ').replace('chu', 'ちゅ').replace('cho', 'ちょ')
    text = text.replace('tya', 'ちゃ').replace('tyu', 'ちゅ').replace('tyo', 'ちょ')
    text = text.replace('nya', 'にゃ').replace('nyu', 'にゅ').replace('nyo', 'にょ')
    text = text.replace('hya', 'ひゃ').replace('hyu', 'ひゅ').replace('hyo', 'ひょ')
    text = text.replace('mya', 'みゃ').replace('myu', 'みゅ').replace('myo', 'みょ')
    text = text.replace('rya', 'りゃ').replace('ryu', 'りゅ').replace('ryo', 'りょ')
    text = text.replace('bya', 'びゃ').replace('byu', 'びゅ').replace('byo', 'びょ')
    text = text.replace('pya', 'ぴゃ').replace('pyu', 'ぴゅ').replace('pyo', 'ぴょ')
    text = text.replace('oh', 'おお')
    text = text.replace('ga', 'が').replace('gi', 'ぎ').replace('gu', 'ぐ')
    text = text.replace('ge', 'げ').replace('go', 'ご').replace('za', 'ざ')
    text = text.replace('ji', 'じ').replace('zu', 'ず').replace('ze', 'ぜ')
    text = text.replace('zo', 'ぞ').replace('da', 'だ').replace('ji', 'ぢ').replace('di', 'ぢ')
    text = text.replace('va', 'ゔぁ').replace('vi', 'ゔぃ').replace('vu', 'ゔ')
    text = text.replace('ve', 'ゔぇ').replace('vo', 'ゔぉ').replace('vya', 'ゔゃ')
    text = text.replace('vyi', 'ゔぃ').replace('vyu', 'ゔゅ').replace('vye', 'ゔぇ')
    text = text.replace('vyo', 'ゔょ')
    text = text.replace('zu', 'づ').replace('de', 'で').replace('do', 'ど')
    text = text.replace('ba', 'ば').replace('bi', 'び').replace('bu', 'ぶ')
    text = text.replace('be', 'べ').replace('bo', 'ぼ').replace('pa', 'ぱ')
    text = text.replace('pi', 'ぴ').replace('pu', 'ぷ').replace('pe', 'ぺ')
    text = text.replace('po', 'ぽ').replace('dha', 'でゃ').replace('dhi', 'でぃ')
    text = text.replace('dhu', 'でゅ').replace('dhe', 'でぇ').replace('dho', 'でょ')
    text = text.replace('ka', 'か').replace('ki', 'き').replace('ku', 'く')
    text = text.replace('ke', 'け').replace('ko', 'こ').replace('sa', 'さ')
    text = text.replace('shi', 'し').replace('su', 'す').replace('se', 'せ')
    text = text.replace('so', 'そ').replace('ta', 'た').replace('chi', 'ち')
    text = text.replace('tsu', 'つ').replace('te', 'て').replace('to', 'と')
    text = text.replace('na', 'な').replace('ni', 'に').replace('nu', 'ぬ')
    text = text.replace('ne', 'ね').replace('no', 'の').replace('ha', 'は')
    text = text.replace('hi', 'ひ').replace('fu', 'ふ').replace('he', 'へ')
    text = text.replace('ho', 'ほ').replace('ma', 'ま').replace('mi', 'み')
    text = text.replace('mu', 'む').replace('me', 'め').replace('mo', 'も')
    text = text.replace('ra', 'ら').replace('ri', 'り').replace('ru', 'る')
    text = text.replace('re', 'れ').replace('ro', 'ろ')
    text = text.replace('ya', 'や').replace('yu', 'ゆ').replace('yo', 'よ')
    text = text.replace('wa', 'わ').replace('wi', 'ゐ').replace('we', 'ゑ')
    text = text.replace('wo', 'を')
    text = text.replace('nn', 'ん').replace('tu', 'つ').replace('hu', 'ふ')
    text = text.replace('fa', 'ふぁ').replace('fi', 'ふぃ').replace('fe', 'ふぇ')
    text = text.replace('fo', 'ふぉ').replace('-', 'ー')
    text = _convert(text, HEP2KANA)
    ret = []
    for (i, char) in enumerate(text):
        if char in consonants:
            char = 'っ'
        ret.append(char)
    return ''.join(ret)



from phonepiece.lang import normalize_lang_id
from phonepiece.lexicon import read_lexicon
from transphone.lang.base_tokenizer import BaseTokenizer
from transphone.lang.eng.tokenizer import read_eng_tokenizer
from transphone.lang.cmn.tokenizer import read_cmn_tokenizer
from transphone.lang.jpn.tokenizer import read_jpn_tokenizer
from transphone.lang.g2p_tokenizer import read_g2p_tokenizer
from transphone.model.utils import resolve_model_name
from transphone.lang.epitran_tokenizer import read_epitran_tokenizer

lang2tokenizer = {
    'eng': read_eng_tokenizer,
    'cmn': read_cmn_tokenizer,
    'jpn': read_jpn_tokenizer,
    'spa': read_epitran_tokenizer,
    'deu': read_epitran_tokenizer,
    'ita': read_epitran_tokenizer,
    'rus': read_epitran_tokenizer,
    'fra': read_epitran_tokenizer,
    'vie': read_epitran_tokenizer,
    'tha': read_epitran_tokenizer,
    'swa': read_epitran_tokenizer,
    'ckb': read_epitran_tokenizer,
    'cat': read_epitran_tokenizer,
}

def read_tokenizer(lang_id, g2p_model='latest', device=None, use_lexicon=True):

    lang_id = normalize_lang_id(lang_id)

    if lang_id in lang2tokenizer:
        return lang2tokenizer[lang_id](lang_id=lang_id, g2p_model=g2p_model, device=device, use_lexicon=use_lexicon)
    else:
        return read_g2p_tokenizer(lang_id=lang_id, g2p_model=g2p_model, device=device)


from transphone.model.dataset import read_zsl_dataset
from transphone.model.utils import read_model_config
from transphone.g2p import read_g2p
import editdistance
import tqdm
from transphone.model.checkpoint_utils import *
from phonepiece.distance import phonological_distance
import argparse


def get_decode_dir(exp_name, checkpoint, ensemble):

    decode_dir = TransphoneConfig.data_path / 'decode' / 'zsl'

    perf = str(checkpoint).split('/')[-1].split('.')[1]

    # create test directory
    exp_dir = decode_dir / f"{exp_name}-{perf}-{ensemble}"
    return exp_dir


def eval_zsl_test(exp, checkpoint, ensemble, device):

    exp_dir = get_decode_dir(exp, checkpoint, ensemble)
    exp_dir.mkdir(exist_ok=True, parents=True)

    test_grapheme_lst, test_phoneme_lst, test_lang_lst = read_zsl_dataset(exp)
    print("test size: ", len(test_lang_lst))

    model = read_g2p(exp, device, checkpoint=checkpoint)

    tot_err = 0
    tot_dst = 0
    tot_csum = 0

    exp_dir = get_decode_dir(exp, checkpoint, ensemble)
    exp_dir.mkdir(exist_ok=True, parents=True)
    log_w = open(exp_dir / 'result.md', 'w')
    log_w.write('| language | phoneme error rate | phonological distance |\n')
    log_w.write('|----------|--------------------|-----------------------|\n')

    for grapheme_lst, phonemes_lst, lang_id in tqdm.tqdm(zip(test_grapheme_lst, test_phoneme_lst, test_lang_lst)):

        err = 0
        dst = 0
        csum = 0

        w = open(exp_dir / f'{lang_id}.txt', 'w')

        for grapheme, phonemes in zip(grapheme_lst, phonemes_lst):

            predicted = model.inference_word_batch(grapheme, lang_id=lang_id, num_lang=ensemble, force_approximate=True)

            w.write(f'lang: {lang_id}\n')
            w.write(f'inp : {grapheme}\n')
            w.write(f'ref : {" ".join(phonemes)}\n')
            w.write(f'hyp : {" ".join(predicted)}\n')
            cur_err = editdistance.distance(phonemes, predicted)
            cur_dst = phonological_distance(phonemes, predicted)
            cur_sum = len(phonemes)
            w.write(f'err: {cur_err / cur_sum}\n')
            w.write(f'dst: {cur_dst / cur_sum}\n\n')
            err += cur_err
            dst += cur_dst
            csum += cur_sum

        w.close()

        tot_err += err
        tot_dst += dst
        tot_csum += csum

        if csum > 0:
            log_w.write(f'| {lang_id} | {err / csum:.3f} | {dst / csum:.3f} |\n')
            print('lang   :', lang_id)
            print("val cer: ", err / csum)
            print("val dst: ",  dst / csum)

    print('all')
    print("val err: ", tot_err / tot_csum)
    print("val dst: ", tot_dst / tot_csum)
    log_w.close()


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='eval zero-shot learning g2p')
    parser.add_argument('--exp', type=str, help='exp')
    parser.add_argument('--checkpoint', type=str, help='checkpoint')
    parser.add_argument('--device', type=str, default='cuda')
    parser.add_argument('--ensemble', type=int, default=10)

    args = parser.parse_args()

    checkpoint = args.checkpoint
    exp = args.exp
    device = args.device
    ensemble = args.ensemble


    if exp is None:
        exp = Path(checkpoint).parent.stem

    if checkpoint is None:
        model_path = TransphoneConfig.data_path / 'model' / exp
        if (model_path / "model.pt").exists():
            checkpoint = model_path / "model.pt"
        else:
            target_model = find_topk_models(model_path)[0]
            print("using model ", target_model)
            checkpoint = target_model

    eval_zsl_test(exp, checkpoint, ensemble, device)






from pathlib import Path
import logging
import torch

class TransphoneConfig:

    root_path = Path(__file__).parent.parent
    data_path = Path(__file__).parent / 'data'
    lang_path = data_path / 'lang'
    logger = logging.getLogger('transphone')
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')



from phonepiece.lexicon import read_lexicon
from phonepiece.lang import read_all_langs
from transphone.model.vocab import Vocab
from transphone.config import TransphoneConfig
from tqdm import tqdm
import torch


def read_dataset(lang_id='eng'):

    lexicon = read_lexicon(lang_id)

    phoneme_lst = []
    grapheme_lst = []

    phoneme_set = set()
    grapheme_set = set()

    for grapheme_str, phonemes in lexicon.word2phoneme.items():
        graphemes = list(grapheme_str)
        phoneme_lst.append(phonemes)
        grapheme_lst.append(graphemes)
        phoneme_set.update(phonemes)
        grapheme_set.update(graphemes)

    phoneme_vocab = Vocab(phoneme_set)
    grapheme_vocab = Vocab(grapheme_set)

    return Dataset(phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab)


def read_multilingual_dataset():

    phoneme_lst = []
    grapheme_lst = []

    test_phoneme_lst = []
    test_grapheme_lst = []

    phoneme_set = set()
    grapheme_set = set()

    print("loading dataset...")

    for lang_id in tqdm(read_all_langs()):

        try:
            lexicon = read_lexicon(lang_id)
        except:
            print("skip ", lang_id)

        if len(lexicon.word2phoneme) < 50:
            continue

        lang_tag = '<'+lang_id+'>'

        grapheme_set.add(lang_tag)

        word2phoneme_lst = list(lexicon.word2phoneme.items())

        # train set
        for grapheme_str, phonemes in word2phoneme_lst[:-50]:
            graphemes = [lang_tag] + list(grapheme_str)
            phoneme_lst.append(phonemes)
            grapheme_lst.append(graphemes)
            phoneme_set.update(phonemes)
            grapheme_set.update(graphemes)

        # dev set
        for grapheme_str, phonemes in word2phoneme_lst[-50:-25]:
            graphemes = [lang_tag] + list(grapheme_str)
            test_phoneme_lst.append(phonemes)
            test_grapheme_lst.append(graphemes)
            phoneme_set.update(phonemes)
            grapheme_set.update(graphemes)

    phoneme_vocab = Vocab(phoneme_set)
    grapheme_vocab = Vocab(grapheme_set)

    train_dataset = Dataset(phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab)
    dev_dataset = Dataset(test_phoneme_lst, test_grapheme_lst, phoneme_vocab, grapheme_vocab)

    return train_dataset, dev_dataset


def read_test_dataset(model_name):

    model_path = TransphoneConfig.data_path / 'model' / model_name
    grapheme_vocab = Vocab.read(model_path / 'grapheme.vocab')
    phoneme_vocab = Vocab.read(model_path / 'phoneme.vocab')

    test_phoneme_lst = []
    test_grapheme_lst = []
    lang_lst = []

    for lang_id in tqdm(read_all_langs()):

        try:
            lexicon = read_lexicon(lang_id)
        except:
            print("skip ", lang_id)

        if len(lexicon.word2phoneme) <= 50:
            continue

        word2phoneme_lst = list(lexicon.word2phoneme.items())
        lang_phoneme_lst = []
        lang_grapheme_lst = []

        # last 25 for testing
        for grapheme_str, phonemes in word2phoneme_lst[-25:]:

            graphemes = list(grapheme_str)

            skip = False

            for phoneme in phonemes:
                if phoneme not in phoneme_vocab:
                    skip = True
                    break

            for grapheme in graphemes:
                if grapheme not in grapheme_vocab:
                    skip = True
                    break

            if skip:
                continue

            lang_phoneme_lst.append(phonemes)
            lang_grapheme_lst.append(grapheme_str)

        if len(lang_phoneme_lst) > 0:
            test_phoneme_lst.append(lang_phoneme_lst)
            test_grapheme_lst.append(lang_grapheme_lst)
            lang_lst.append(lang_id)

    return test_grapheme_lst, test_phoneme_lst, lang_lst


def read_zsl_dataset(model_name):

    model_path = TransphoneConfig.data_path / 'model' / model_name
    grapheme_vocab = Vocab.read(model_path / 'grapheme.vocab')
    phoneme_vocab = Vocab.read(model_path / 'phoneme.vocab')

    test_phoneme_lst = []
    test_grapheme_lst = []
    lang_lst = []

    for lang_id in tqdm(read_all_langs()):

        try:
            lexicon = read_lexicon(lang_id)
        except:
            print("skip ", lang_id)

        if len(lexicon.word2phoneme) >= 50 or len(lexicon.word2phoneme) == 0:
            continue

        word2phoneme_lst = list(lexicon.word2phoneme.items())

        lang_phoneme_lst = []
        lang_grapheme_lst = []

        # last 10 for testing
        for grapheme_str, phonemes in word2phoneme_lst[-25:]:
            graphemes = list(grapheme_str)

            skip = False

            for phoneme in phonemes:
                if phoneme not in phoneme_vocab:
                    skip = True
                    break

            for grapheme in graphemes:
                if grapheme not in grapheme_vocab:
                    skip = True
                    break

            if skip:
                continue

            lang_phoneme_lst.append(phonemes)
            lang_grapheme_lst.append(grapheme_str)

        test_phoneme_lst.append(lang_phoneme_lst)
        test_grapheme_lst.append(lang_grapheme_lst)
        lang_lst.append(lang_id)

    return test_grapheme_lst, test_phoneme_lst, lang_lst


class Dataset:

    def __init__(self, phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab):

        self.phoneme_lst = phoneme_lst
        self.grapheme_lst = grapheme_lst
        self.phoneme_vocab = phoneme_vocab
        self.grapheme_vocab = grapheme_vocab

    def __getitem__(self, item):

        phones = self.phoneme_lst[item]
        graphemes = self.grapheme_lst[item]

        phone_ids = [self.phoneme_vocab.atoi(phone) for phone in phones]
        grapheme_ids = [self.grapheme_vocab.atoi(grapheme) for grapheme in graphemes]

        return (torch.LongTensor(grapheme_ids), torch.LongTensor(phone_ids))

    def __len__(self):
        return len(self.phoneme_lst)


class P2GDataset:

    def __init__(self, phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab):

        self.phoneme_lst = phoneme_lst
        self.grapheme_lst = grapheme_lst
        self.phoneme_vocab = phoneme_vocab
        self.grapheme_vocab = grapheme_vocab

    def __getitem__(self, item):

        phones = self.phoneme_lst[item]
        graphemes = self.grapheme_lst[item]

        phone_ids = [self.phoneme_vocab.atoi(phone) for phone in phones]
        grapheme_ids = [self.grapheme_vocab.atoi(grapheme) for grapheme in graphemes]

        return (torch.LongTensor(phone_ids), torch.LongTensor(grapheme_ids))

    def __len__(self):
        return len(self.phoneme_lst)


from phonepiece.lang import normalize_lang_id
from transphone.model.checkpoint_utils import torch_load
from transphone.model.transformer import TransformerG2P
from transphone.model.ensemble import ensemble
from transphone.bin.download_model import download_model
from transphone.config import TransphoneConfig
from transphone.model.checkpoint_utils import find_topk_models
from transphone.model.vocab import Vocab
from transphone.model.utils import read_model_config
from transphone.utils import Singleton
from phonepiece.lang import read_tree
from phonepiece.inventory import read_inventory
from pathlib import Path
import torch
import unidecode
from itertools import groupby
from transphone.model.utils import resolve_model_name


def read_g2p(model_name='latest', device=None, checkpoint=None):

    if device is not None:
        assert device in ['cpu', 'cuda']
        TransphoneConfig.device = device

    model_name = resolve_model_name(model_name)
    cache_path = None

    if checkpoint is None:
        model_path = TransphoneConfig.data_path / 'model' / model_name

        # if not exists, we try to download the model
        if not model_path.exists():
            download_model(model_name)

        if not model_path.exists():
            raise ValueError(f"could not download or read {model_name} model")

        if (model_path / "model.pt").exists():
            checkpoint = model_path / "model.pt"
        else:
            checkpoint = find_topk_models(model_path)[0]

        if (model_path / 'cache').exists():
            cache_path = model_path / 'cache'
        else:
            cache_path = None

    config = read_model_config(model_name)

    model = G2P(checkpoint, cache_path, config)

    return model


class G2P(metaclass=Singleton):

    def __init__(self, checkpoint, cache_path, config):

        self.model_path = Path(checkpoint).parent
        self.grapheme_vocab = Vocab.read(self.model_path / 'grapheme.vocab')
        self.phoneme_vocab = Vocab.read(self.model_path / 'phoneme.vocab')
        self.config = config
        self.checkpoint = checkpoint
        self.cache_path = cache_path

        # setup available languages
        self.supervised_langs = []
        for word in self.grapheme_vocab.words[2:]:
            if len(word) == 5 and word[0] == '<' and word[-1] == '>':
                self.supervised_langs.append(word[1:-1])

        # cache to find proper supervised language
        self.lang_map = {}

        # inventory
        self.lang2inv = {}

        # lang2cache
        self.lang2cache = {}

        # tree to estimate language's similarity
        self.tree = read_tree()
        self.tree.setup_target_langs(self.supervised_langs)
        self.supervised_langs = set(self.supervised_langs)

        SRC_VOCAB_SIZE = len(self.grapheme_vocab)+1
        TGT_VOCAB_SIZE = len(self.phoneme_vocab)+1

        EMB_SIZE = config.embed_size
        NHEAD = config.num_head
        FFN_HID_DIM = config.hidden_size
        NUM_ENCODER_LAYERS = config.num_encoder
        NUM_DECODER_LAYERS = config.num_decoder

        torch.manual_seed(0)

        self.model = TransformerG2P(NUM_ENCODER_LAYERS, NUM_DECODER_LAYERS, EMB_SIZE,
                            NHEAD, SRC_VOCAB_SIZE, TGT_VOCAB_SIZE, FFN_HID_DIM).to(TransphoneConfig.device)

        torch_load(self.model, self.checkpoint)


    def get_target_langs(self, lang_id, num_lang=10, verbose=False, force_approximate=False):

        if lang_id in self.lang_map:
            target_langs = self.lang_map[lang_id]
        else:

            if force_approximate or lang_id not in self.supervised_langs:
                target_langs = self.tree.get_nearest_langs(lang_id, num_lang)
                if verbose:
                    print("lang ", lang_id, " is not available directly, use ", target_langs, " instead")
                self.lang_map[lang_id] = target_langs
            else:
                self.lang_map[lang_id] = [lang_id]
                target_langs = [lang_id]

        return target_langs

    def inference_word(self, word, lang_id='eng', num_lang=10, verbose=False, force_approximate=False):

        target_langs = self.get_target_langs(lang_id, num_lang, verbose, force_approximate)

        phones_lst = []

        for target_lang_id in target_langs:
            lang_tag = '<' + target_lang_id + '>'

            graphemes = [lang_tag]+[w.lower() for w in list(word)]

            grapheme_ids = []
            for grapheme in graphemes:
                if grapheme not in self.grapheme_vocab:

                    # romanize chars not available in training languages
                    romans = list(unidecode.unidecode(grapheme))

                    if verbose:
                        print("WARNING: not found grapheme ", grapheme, " in vocab. use ", romans, " instead")

                    for roman in romans:

                        # discard special chars such as $
                        if roman in self.grapheme_vocab:
                            grapheme_ids.append(self.grapheme_vocab.atoi(roman))
                    continue
                grapheme_ids.append(self.grapheme_vocab.atoi(grapheme))

            x = torch.LongTensor(grapheme_ids).unsqueeze(0).to(TransphoneConfig.device)

            phone_ids = self.model.inference(x)

            phones = [self.phoneme_vocab.itoa(phone) for phone in phone_ids]

            # ignore empty
            if len(phones) == 0:
                continue

            # if it is a mapped language, we need to map the inference_phone to the correct language inventory
            if lang_id not in self.lang2inv:
                inv = read_inventory(lang_id)
                self.lang2inv[lang_id] = inv

            inv = self.lang2inv[lang_id]
            phones = inv.remap(phones)

            if verbose:
                print(target_lang_id, ' ', phones)

            phones_lst.append(phones)


        if len(phones_lst) == 0:
            phones = []
        else:
            phones = ensemble(phones_lst)

        return phones

    def inference_word_batch(self, word, lang_id='eng', num_lang=10, verbose=False, force_approximate=False):

        target_langs = self.get_target_langs(lang_id, num_lang, verbose, force_approximate)

        phones_lst = []

        grapheme_input = []

        for target_lang_id in target_langs:
            lang_tag = '<' + target_lang_id + '>'

            graphemes = [lang_tag]+[w.lower() for w in list(word)]

            grapheme_ids = []
            normalized_graphemes = []

            for grapheme in graphemes:
                if grapheme not in self.grapheme_vocab:

                    # romanize chars not available in training languages
                    romans = list(unidecode.unidecode(grapheme))

                    if verbose:
                        print("WARNING: not found grapheme ", grapheme, " in vocab. use ", romans, " instead")

                    for roman in romans:

                        # discard special chars such as $
                        if roman in self.grapheme_vocab:
                            grapheme_ids.append(self.grapheme_vocab.atoi(roman))
                            normalized_graphemes.append(roman)
                    continue

                normalized_graphemes.append(grapheme)
                grapheme_ids.append(self.grapheme_vocab.atoi(grapheme))

            grapheme_input.append(grapheme_ids)

            if verbose:
                print(f"normalized: {word} -> {normalized_graphemes}")

        x = torch.LongTensor(grapheme_input).to(TransphoneConfig.device)

        phone_output = self.model.inference_batch(x)

        for target_lang_id, phone_ids in zip(target_langs, phone_output):
            phones = [self.phoneme_vocab.itoa(phone) for phone in phone_ids]

            # ignore empty
            if len(phones) == 0:
                continue

            # if it is a mapped language, we need to map the inference_phone to the correct language inventory
            if lang_id not in self.lang2inv:
                inv = read_inventory(lang_id)
                self.lang2inv[lang_id] = inv

            inv = self.lang2inv[lang_id]
            phones = inv.remap(phones)

            if verbose:
                print(target_lang_id, ' ', phones)

            phones_lst.append(phones)

        if len(phones_lst) == 0:
            phones = []
        else:
            phones = ensemble(phones_lst)

        return phones

    def inference(self, text, lang_id='eng', num_lang=10, verbose=False, force_approximate=False):
        lang_id = normalize_lang_id(lang_id)

        phones_lst = []

        words = text.split()

        for word in words:
            phones = self.inference_word(word, lang_id, num_lang, verbose, force_approximate)
            phones = [x[0] for x in groupby(phones)]
            phones_lst.extend(phones)

        return phones_lst

    def inference_batch(self, text, lang_id='eng', num_lang=10, verbose=False, force_approximate=False):
        lang_id = normalize_lang_id(lang_id)

        phones_lst = []

        words = text.split()

        for word in words:
            phones = self.inference_word_batch(word, lang_id, num_lang, verbose, force_approximate)
            phones = [x[0] for x in groupby(phones)]
            phones_lst.extend(phones)

        return phones_lst


import numpy as np
from collections import defaultdict

def test():

    a = Lattice("abcd")
    b = Lattice("bzde")
    c = combine(a,b)
    d = Lattice("bcdef")
    e = combine(c, d)

    print(e)

def ensemble(pred_lst):

    assert len(pred_lst) >= 1

    lattice_base = None

    for pred in pred_lst:
        if not isinstance(pred, list):
            pred = pred.split()

        if lattice_base is None:
            lattice_base = Lattice(pred)

        else:
            lattice_base = combine(lattice_base, Lattice(pred))

    return lattice_base.compute()


def combine(lattice_a, lattice_b, verbose=True, out=None):

    # output
    lattice = Lattice()

    cs_lst = []

    # length of each string
    len_a = len(lattice_a)
    len_b = len(lattice_b)

    # dp table
    dp = [[0 for x in range(len_a+1)] for y in range(len_b+1)]
    path = [[(0, 0) for x in range(len_a+1)] for y in range(len_b+1)]

    # initialize first row and first column
    for i in range(1, len_a+1):
        dp[0][i] = i
        path[0][i] = (0, i-1)

    for i in range(1, len_b+1):
        dp[i][0] = i
        path[i][0] = (i-1, 0)

    # dp update
    for i in range(1, len_b+1):
        for j in range(1, len_a+1):
            index_a = j-1
            index_b = i-1

            cs_a = lattice_a[index_a]
            cs_b = lattice_b[index_b]

            sub_cost = cs_a.substitute_cost(cs_b)
            del_cost = cs_a.delete_cost()
            add_cost = cs_b.delete_cost()

            dp[i][j] = dp[i-1][j-1]+sub_cost
            path[i][j] = (i-1,j-1)

            if dp[i][j] > dp[i-1][j]+del_cost:
                dp[i][j] = dp[i-1][j]+del_cost
                path[i][j] = (i-1,j)

            if dp[i][j] > dp[i][j-1]+add_cost:
                dp[i][j] = dp[i][j-1]+add_cost
                path[i][j] = (i, j-1)

    cur_node = (len_b, len_a)

    while(cur_node != (0,0)):
        prev_node = path[cur_node[0]][cur_node[1]]

        cs_a = lattice_a[cur_node[1]-1]
        cs_b = lattice_b[cur_node[0]-1]

        # substitution or match
        if prev_node[0]+1 == cur_node[0] and prev_node[1]+1 == cur_node[1]:

            cs_a.merge(cs_b)
            cs_lst.append(cs_a)


        # addition
        if prev_node[0] + 1 == cur_node[0] and prev_node[1] == cur_node[1]:

            cs_e = cs_a.create_empty_set()
            cs_e.merge(cs_b)

            cs_lst.append(cs_e)

        # deletion
        if prev_node[0] == cur_node[0] and prev_node[1]+1 == cur_node[1]:

            cs_e = cs_b.create_empty_set()
            cs_e.merge(cs_a)
            cs_lst.append(cs_e)

        cur_node = prev_node

    cs_lst.reverse()
    lattice.cs_lst = cs_lst

    return lattice

class CorrespondenceSet:

    def __init__(self, units=None, scores=None):

        if units is None:
            units = []

        if scores is None:
            scores = [1.0]*len(units)

        self.units = units
        self.unit_set = set(units)
        self.scores = scores


    def __repr__(self):
        return '{'+','.join(self.units)+'}'

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        for unit in self.units:
            yield unit

    def __len__(self):
        return len(self.units)

    def __getitem__(self, idx):
        return self.units[idx]

    def compute(self):
        unit2score = defaultdict(float)

        for i in range(len(self.units)):
            unit = self.units[i]
            score = self.scores[i]
            unit2score[unit] += score

        return sorted(unit2score.items(), key=lambda x:-x[1])[0][0]

    def has_epsilon(self):
        return '<blk>' in self.unit_set

    def contains(self, unit):
        return unit in self.unit_set

    def delete_cost(self):
        if self.has_epsilon():
            return 0
        else:
            return 1

    def substitute_cost(self, other):
        overlap = False
        for unit in self.units:
            if other.contains(unit):
                overlap = True

        if overlap:
            return 0
        else:
            return 1

    def create_empty_set(self):
        ave_score = np.mean(self.scores)
        cs = CorrespondenceSet(['<blk>'], [ave_score])
        return cs

    def merge(self, other):

        for unit, score in zip(other.units, other.scores):
            self.units.append(unit)
            self.scores.append(score)
            self.unit_set.add(unit)


class Lattice:

    def __init__(self, units=None, scores=None):

        if units is None:
            units = []
        else:
            units = list(units)

        if scores is None:
            scores = [1.0]*len(units)
        else:
            scores = list(scores)

        assert len(units) == len(scores)

        self.cs_lst = []

        for unit, score in zip(units, scores):
            self.cs_lst.append(CorrespondenceSet([unit], [score]))

    def __repr__(self):
        return '\n'.join([str(cs) for cs in self.cs_lst])

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self.cs_lst)

    def __getitem__(self, idx):
        return self.cs_lst[idx]

    def __iter__(self):
        for cs in self.cs_lst:
            yield cs

    def compute(self):
        cs_items = [cs.compute() for cs in self.cs_lst]

        return [cs_item for cs_item in cs_items if cs_item != "<blk>" and not cs_item.isspace()]


import pip
import importlib

def import_with_auto_install(package, package_name):
    try:
        return importlib.import_module(package)
    except ImportError:
        pip.main(['install', package_name])
    return importlib.import_module(package)



class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


from .conv_table import FULL_KANA
from phonepiece.inventory import read_inventory
from transphone.config import TransphoneConfig

_kana2phonemes = {
    'ア': 'a',
    'イ': 'i',
    'ウ': 'u',
    'エ': 'e',
    'オ': 'o',
    'カ': 'k a',
    'キ': 'k i',
    'ク': 'k u',
    'ケ': 'k e',
    'コ': 'k o',
    'ガ': 'g a',
    'ギ': 'g i',
    'グ': 'g u',
    'ゲ': 'g e',
    'ゴ': 'g o',
    'サ': 's a',
    'シ': 'sh i',
    'ス': 's u',
    'セ': 's e',
    'ソ': 's o',
    'ザ': 'z a',
    'ジ': 'j i',
    'ズ': 'z u',
    'ゼ': 'z e',
    'ゾ': 'z o',
    'タ': 't a',
    'チ': 'ch i',
    'ツ': 't͡s u',
    'テ': 't e',
    'ト': 't o',
    'ダ': 'd a',
    'ヂ': 'j i',
    'ヅ': 'z u',
    'デ': 'd e',
    'ド': 'd o',
    'ナ': 'n a',
    'ニ': 'n i',
    'ヌ': 'n u',
    'ネ': 'n e',
    'ノ': 'n o',
    'ハ': 'h a',
    'ヒ': 'h i',
    'フ': 'f u',
    'ヘ': 'h e',
    'ホ': 'h o',
    'バ': 'b a',
    'ビ': 'b i',
    'ブ': 'b u',
    'ベ': 'b e',
    'ボ': 'b o',
    'パ': 'p a',
    'ピ': 'p i',
    'プ': 'p u',
    'ペ': 'p e',
    'ポ': 'p o',
    'マ': 'm a',
    'ミ': 'm i',
    'ム': 'm u',
    'メ': 'm e',
    'モ': 'm o',
    'ラ': 'r a',
    'リ': 'r i',
    'ル': 'r u',
    'レ': 'r e',
    'ロ': 'r o',
    'ワ': 'w a',
    'ヲ': 'o',
    'ヤ': 'y a',
    'ユ': 'y u',
    'ヨ': 'y o',
    'キャ': 'ky a',
    'キュ': 'ky u',
    'キョ': 'ky o',
    'ギャ': 'gy a',
    'ギュ': 'gy u',
    'ギョ': 'gy o',
    'シャ': 'sh a',
    'シュ': 'sh u',
    'ショ': 'sh o',
    'ジャ': 'j a',
    'ジュ': 'j u',
    'ジョ': 'j o',
    'チャ': 'ch a',
    'チュ': 'ch u',
    'チョ': 'ch o',
    'ニャ': 'ny a',
    'ニュ': 'ny u',
    'ニョ': 'ny o',
    'ヒャ': 'hy a',
    'ヒュ': 'hy u',
    'ヒョ': 'hy o',
    'ビャ': 'by a',
    'ビュ': 'by u',
    'ビョ': 'by o',
    'ピャ': 'py a',
    'ピュ': 'py u',
    'ピョ': 'py o',
    'ミャ': 'my a',
    'ミュ': 'my u',
    'ミョ': 'my o',
    'リャ': 'ry a',
    'リュ': 'ry u',
    'リョ': 'ry o',
    'イェ': 'i e',
    'シェ': 'sh e',
    'ジェ': 'j e',
    'ティ': 't i',
    'トゥ': 't u',
    'チェ': 'ch e',
    'ツァ': 't͡s a',
    'ツィ': 't͡s i',
    'ツェ': 't͡s e',
    'ツォ': 't͡s o',
    'ディ': 'd i',
    'ドゥ': 'd u',
    'デュ': 'd u',
    'ニェ': 'n i e',
    'ヒェ': 'h e',
    'ファ': 'f a',
    'フィ': 'f i',
    'フェ': 'f e',
    'フォ': 'f o',
    'フュ': 'hy u',
    'ブィ': 'b i',
    'ミェ': 'm e',
    'ウィ': 'w i',
    'ウェ': 'w e',
    'ウォ': 'w o',
    'クヮ': 'k a',
    'グヮ': 'g a',
    'スィ': 's u i',
    'ズィ': 'j i',
    'テュ': 't e y u',
    'ヴァ': 'b a',
    'ヴィ': 'b i',
    'ヴ': 'b u',
    'ヴェ': 'b e',
    'ヴォ': 'b o',
    'ン': 'N',
    'ッ': 'q',
    'ー': 'ː'
}

import re

class Kana2Phoneme:
    def __init__(self):

        self.phoneme_set = set(read_inventory('jpn').phoneme.elems)

        self._dict1 = {
            'キャ': 'ky a ',
            'キュ': 'ky u ',
            'キョ': 'ky o ',
            'ギャ': 'gy a ',
            'ギュ': 'gy u ',
            'ギョ': 'gy o ',
            'シャ': 'sh a ',
            'シュ': 'sh u ',
            'ショ': 'sh o ',
            'ジャ': 'j a ',
            'ジュ': 'j u ',
            'ジョ': 'j o ',
            'チャ': 'ch a ',
            'チュ': 'ch u ',
            'チョ': 'ch o ',
            'ニャ': 'ny a ',
            'ニュ': 'ny u ',
            'ニョ': 'ny o ',
            'ヒャ': 'hy a ',
            'ヒュ': 'hy u ',
            'ヒョ': 'hy o ',
            'ビャ': 'by a ',
            'ビュ': 'by u ',
            'ビョ': 'by o ',
            'ピャ': 'py a ',
            'ピュ': 'py u ',
            'ピョ': 'py o ',
            'ミャ': 'my a ',
            'ミュ': 'my u ',
            'ミョ': 'my o ',
            'リャ': 'ry a ',
            'リュ': 'ry u ',
            'リョ': 'ry o ',
            'イェ': 'i e ',
            'シェ': 'sh e ',
            'ジェ': 'j e ',
            'ティ': 't i ',
            'トゥ': 't u ',
            'チェ': 'ch e ',
            'ツァ': 't͡s a ',
            'ツィ': 't͡s i ',
            'ツェ': 't͡s e ',
            'ツォ': 't͡s o ',
            'ディ': 'd i ',
            'ドゥ': 'd u ',
            'デュ': 'd u ',
            'ニェ': 'n i e ',
            'ヒェ': 'h e ',
            'ファ': 'f a ',
            'フィ': 'f i ',
            'フェ': 'f e ',
            'フォ': 'f o ',
            'フュ': 'hy u ',
            'ブィ': 'b i ',
            'ミェ': 'm e ',
            'ウィ': 'w i ',
            'ウェ': 'w e ',
            'ウォ': 'w o ',
            'クヮ': 'k a ',
            'グヮ': 'g a ',
            'スィ': 's u i ',
            'ズィ': 'j i ',
            'テュ': 't e y u ',
            'ヴァ': 'b a ',
            'ヴィ': 'b i ',
            'ヴ': 'b u ',
            'ヴェ': 'b e ',
            'ヴォ': 'b o ',
        }
        self._dict2 = {
            'ア': 'a ',
            'イ': 'i ',
            'ウ': 'u ',
            'エ': 'e ',
            'オ': 'o ',
            'カ': 'k a ',
            'キ': 'k i ',
            'ク': 'k u ',
            'ケ': 'k e ',
            'コ': 'k o ',
            'ガ': 'g a ',
            'ギ': 'g i ',
            'グ': 'g u ',
            'ゲ': 'g e ',
            'ゴ': 'g o ',
            'サ': 's a ',
            'シ': 'sh i ',
            'ス': 's u ',
            'セ': 's e ',
            'ソ': 's o ',
            'ザ': 'z a ',
            'ジ': 'j i ',
            'ズ': 'z u ',
            'ゼ': 'z e ',
            'ゾ': 'z o ',
            'タ': 't a ',
            'チ': 'ch i ',
            'ツ': 't͡s u ',
            'テ': 't e ',
            'ト': 't o ',
            'ダ': 'd a ',
            'ヂ': 'j i ',
            'ヅ': 'z u ',
            'デ': 'd e ',
            'ド': 'd o ',
            'ナ': 'n a ',
            'ニ': 'n i ',
            'ヌ': 'n u ',
            'ネ': 'n e ',
            'ノ': 'n o ',
            'ハ': 'h a ',
            'ヒ': 'h i ',
            'フ': 'f u ',
            'ヘ': 'h e ',
            'ホ': 'h o ',
            'バ': 'b a ',
            'ビ': 'b i ',
            'ブ': 'b u ',
            'ベ': 'b e ',
            'ボ': 'b o ',
            'パ': 'p a ',
            'ピ': 'p i ',
            'プ': 'p u ',
            'ペ': 'p e ',
            'ポ': 'p o ',
            'マ': 'm a ',
            'ミ': 'm i ',
            'ム': 'm u ',
            'メ': 'm e ',
            'モ': 'm o ',
            'ラ': 'r a ',
            'リ': 'r i ',
            'ル': 'r u ',
            'レ': 'r e ',
            'ロ': 'r o ',
            'ワ': 'w a ',
            'ヲ': 'o ',
            'ヤ': 'y a ',
            'ユ': 'y u ',
            'ヨ': 'y o ',
            'ン': 'N ',
            'ッ': 'q ',
            'ー': 'ː ',
            'ァ': 'a ',
            'ィ': 'i ',
            'ゥ': 'u ',
            'ェ': 'e ',
            'ォ': 'o ',
            'ャ': 'y a ',
            'ュ': 'y u ',
            'ョ': 'y o ',
            'ヮ': 'w a ',
            'ヵ': 'k a ',
            'ヶ': 'k e ',
            'ヰ': 'i ',
            'ヱ': 'e ',
            'ヴ': 'b u ',
            'ヽ': '',  # this should not reach here though
            'ヾ': '',  # this should not reach here though
            'ヷ': 'w a ',
            'ヸ': 'i ',
            'ヹ': 'e ',
            'ヺ': 'o ',
            'ヿ': '',  # this should not reach here though
            '゛': '',  # this should not reach here though
            '゜': '',  # this should not reach here though
            'ヽ': '',  # this should not reach here though
            'ヾ': '',  # this should not reach here though
            'ゝ': '',  # this should not reach here though
            'ゞ': '',  # this should not reach here though
            '〆': '',  # this should not reach here though
            '々': '',  # this should not reach here though
        }
        self._regex1 = re.compile(u"(%s)" % u"|".join(map(re.escape, self._dict1.keys())))
        self._regex2 = re.compile(u"(%s)" % u"|".join(map(re.escape, self._dict2.keys())))

    def validate(self, text):
        for word in text.strip():
            if word not in FULL_KANA:
                return False
        return True

    def convert(self, origin_text):

        if isinstance(origin_text, list):
            text = ' '.join(origin_text)
        else:
            text = origin_text

        ret = text
        ret = self._regex1.sub(lambda m: self._dict1[m.string[m.start():m.end()]], ret)
        ret = self._regex2.sub(lambda m: self._dict2[m.string[m.start():m.end()]], ret)

        temp_phonemes = ret.split()

        phonemes = []
        for temp_phoneme in temp_phonemes:
            if temp_phoneme == 'ː':
                if len(phonemes) > 0 and phonemes[-1] in ['a', 'i', 'u', 'e', 'o']:
                    phonemes[-1] = phonemes[-1]+'ː'
                continue

            if temp_phoneme in self.phoneme_set:
                phonemes.append(temp_phoneme)
            else:
                TransphoneConfig.logger.error("Unknown phoneme: %s" % temp_phoneme)

        return phonemes



from transphone.utils import import_with_auto_install
from transphone.lang.base_tokenizer import BaseTokenizer
from phonepiece.pinyin import PinyinConverter
from transphone.lang.cmn.normalizer import CMNNormalizer


def read_cmn_tokenizer(lang_id='cmn', g2p_model='latest', device=None, use_lexicon=True):
    return CMNTokenizer(lang_id, g2p_model, device)


class CMNTokenizer(BaseTokenizer):

    def __init__(self, lang_id='cmn', g2p_model='latest', device=None):

        super().__init__(lang_id, g2p_model, device)

        # import jieba and pypinyin for segmentation
        self.jieba = import_with_auto_install('jieba', 'jieba')
        pypinyin = import_with_auto_install('pypinyin', 'pypinyin')

        self.pinyin = pypinyin.lazy_pinyin
        self.converter = PinyinConverter()
        self.normalizer = CMNNormalizer()

    def tokenize_words(self, text):

        text = self.normalizer(text)
        words = list(self.jieba.cut(text, use_paddle=False))
        return words


    def tokenize(self, text, use_g2p=True, use_space=False, verbose=False):

        norm_text = self.normalizer(text)

        log = f"normalization: {text} -> {norm_text}"
        self.logger.info(log)
        if verbose:
            print(log)

        text = norm_text

        words = list(self.jieba.cut(text, use_paddle=False))

        ipa_lst = []

        for word in words:
            pinyins = self.pinyin(word)

            self.logger.info(f"pinyin: {word} -> {pinyins}")
            if verbose:
                print(f"pinyin: {word} -> {pinyins}")

            for pinyin in pinyins:
                ipa_lst.extend(self.converter.convert(pinyin))
            if use_space:
                ipa_lst.append('<SPACE>')
        return ipa_lst


# encoding: utf8
"""
from: https://github.com/neologd/mecab-ipadic-neologd/wiki/Regexp.ja
"""
from __future__ import unicode_literals
import re
import unicodedata

def unicode_normalize(cls, s):
    pt = re.compile('([{}]+)'.format(cls))

    def norm(c):
        return unicodedata.normalize('NFKC', c) if pt.match(c) else c

    s = ''.join(norm(x) for x in re.split(pt, s))
    s = re.sub('－', '-', s)
    return s

def remove_extra_spaces(s):
    s = re.sub('[ 　]+', ' ', s)
    blocks = ''.join(('\u4E00-\u9FFF',  # CJK UNIFIED IDEOGRAPHS
                      '\u3040-\u309F',  # HIRAGANA
                      '\u30A0-\u30FF',  # KATAKANA
                      '\u3000-\u303F',  # CJK SYMBOLS AND PUNCTUATION
                      '\uFF00-\uFFEF'   # HALFWIDTH AND FULLWIDTH FORMS
                      ))
    basic_latin = '\u0000-\u007F'

    def remove_space_between(cls1, cls2, s):
        p = re.compile('([{}]) ([{}])'.format(cls1, cls2))
        while p.search(s):
            s = p.sub(r'\1\2', s)
        return s

    s = remove_space_between(blocks, blocks, s)
    s = remove_space_between(blocks, basic_latin, s)
    s = remove_space_between(basic_latin, blocks, s)
    return s

def normalize_neologd(s):
    s = s.strip()
    s = unicode_normalize('０-９Ａ-Ｚａ-ｚ｡-ﾟ', s)

    def maketrans(f, t):
        return {ord(x): ord(y) for x, y in zip(f, t)}

    s = re.sub('[˗֊‐‑‒–⁃⁻₋−]+', '-', s)  # normalize hyphens
    s = re.sub('[﹣－ｰ—―─━ー]+', 'ー', s)  # normalize choonpus
    s = re.sub('[~∼∾〜〰～]', '', s)  # remove tildes
    s = s.translate(
        maketrans('!"#$%&\'()*+,-./:;<=>?@[¥]^_`{|}~｡､･｢｣',
              '！”＃＄％＆’（）＊＋，－．／：；＜＝＞？＠［￥］＾＿｀｛｜｝〜。、・「」'))

    s = remove_extra_spaces(s)
    s = unicode_normalize('！”＃＄％＆’（）＊＋，－．／：；＜＞？＠［￥］＾＿｀｛｜｝〜', s)  # keep ＝,・,「,」
    s = re.sub('[’]', '\'', s)
    s = re.sub('[”]', '"', s)
    return s


def parse_small_jpn_number(num):

    if num == 0:
        return "ぜろ"
    if num == 1:
        return "いち"

    digit = ["", "いち", "に", "さん", "よん", "ご", "ろく", "なな", "はち", "きゅう"]
    unit = ["", "じゅう", "ひゃく", "せん"]
    dakuon_unit = ["", "じゅう", "びゃく", "ぜん"]
    sokuon_unit = ["", "じゅう", "ぴゃく", "せん"]

    num_str = str(num)
    n = len(num_str)
    read_list = []
    for i in range(n):
        d = int(num_str[i])
        dakuon = False
        sokuon = False
        if d != 0:
            # skip 1 unless it is the last char
            if i != n-1 and d == 1:
                read_list.append("")
            # 濁音
            elif d == 3:
                dakuon = True
                read_list.append(digit[d])
            elif d == 6 and n-i == 3:
                read_list.append("ろっ")
                sokuon = True
            elif d == 8 and n-i == 3:
                read_list.append("はっ")
                sokuon = True
            else:
                read_list.append(digit[d])

            if dakuon:
                read_list.append(dakuon_unit[n-i-1])
            elif sokuon:
                read_list.append(sokuon_unit[n-i-1])
            else:
                read_list.append(unit[n - i - 1])
    read_str = "".join(read_list)
    return read_str


def parse_jpn_alphabet(text):

    katakana_dict = {
        'A': 'エー',
        'B': 'ビー',
        'C': 'シー',
        'D': 'ディー',
        'E': 'イー',
        'F': 'エフ',
        'G': 'ジー',
        'H': 'エイチ',
        'I': 'アイ',
        'J': 'ジェー',
        'K': 'ケー',
        'L': 'エル',
        'M': 'エム',
        'N': 'エヌ',
        'O': 'オー',
        'P': 'ピー',
        'Q': 'キュー',
        'R': 'アール',
        'S': 'エス',
        'T': 'ティー',
        'U': 'ユー',
        'V': 'ブイ',
        'W': 'ダブリュー',
        'X': 'エックス',
        'Y': 'ワイ',
        'Z': 'ゼット',
    }

    katakana_text = ''
    for char in text.upper():
        if char in katakana_dict:
            katakana_text += katakana_dict[char]
    return katakana_text

def parse_jpn_number(num):

    num_size = len(str(num))
    if num_size <= 4:
        return parse_small_jpn_number(num)
    elif num_size <= 8:
        low_digit = num%10000
        high_digit = num//10000

        high_read = parse_small_jpn_number(high_digit) + 'まん'
        low_read = ""
        if low_digit != 0:
            low_read = parse_small_jpn_number(low_digit)
        return high_read + low_read
    else:

        low_digit = num % 10000
        mid_digit = (num//10000)%10000
        high_digit = num//100000000

        high_read = parse_small_jpn_number(high_digit) + 'おく'
        mid_read = ""
        if mid_digit != 0:
            mid_read = parse_small_jpn_number(mid_digit) + 'まん'
        low_read = ""
        if low_digit != 0:
            low_read = parse_small_jpn_number(low_digit)

        return high_read + mid_read + low_read


from pathlib import Path


class Vocab:

    def __init__(self, word_set):
        self.words = ['<pad>', '<eos>']+list(sorted(word_set))
        self.map = dict()

        for i, word in enumerate(self.words):
            self.map[word] = i


    @classmethod
    def read(cls, file_path):

        vocab = cls([])

        vocab.words = []
        vocab.map = dict()

        for i, line in enumerate(open(Path(file_path), encoding='utf-8')):
            word = line.strip()
            vocab.words.append(word)
            vocab.map[word] = i

        return vocab

    def __len__(self):
        return len(self.words)

    def __contains__(self, item):
        return item in self.map

    def atoi(self, word):

        return self.map[word]

    def itoa(self, idx):
        word = self.words[idx]
        if word == '<space>':
            word = ' '

        return word

    def write(self, file_path):

        w = open(file_path, 'w', encoding='utf-8')

        for word in self.words:
            w.write(word+'\n')

        w.close()


from pathlib import Path
from transphone.config import TransphoneConfig
import shutil
import yaml


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def read_model_config(exp):
    yaml_file = TransphoneConfig.data_path / 'exp' / f'{exp}.yml'
    # Open the YAML file and load its contents into a Python dictionary
    with open(yaml_file, "r") as f:
        model_config = yaml.safe_load(f)

    return dotdict(model_config)


def resolve_model_name(model_name='latest', alt_model_path=None):
    """
    select the model

    :param model_name:
    :return:
    """

    models = {
        'latest': '042801_base',
        '042801_base': '042801_base'
    }

    assert model_name in models, f"{model_name} is not available"

    return models[model_name]


def pad_list(tensor_lst):
    max_length = max(t.size(0) for t in tensor_lst)
    batch_size = len(tensor_lst)

    padded_tensor = tensor_lst[0].new(batch_size, max_length, *tensor_lst[0].size()[1:]).fill_(0)
    for i, t in enumerate(tensor_lst):
        padded_tensor[i,:t.size(0)] = t

    return padded_tensor


def pad_sos_eos(ys, sos, eos):

    batch_size = len(ys)

    sos_tensor = ys.new_zeros((batch_size, 1)).fill_(sos)
    y_in = torch.cat([sos_tensor, ys], dim=1)

    y_out = []

    zero_tensor = ys.new_zeros((batch_size, 1))
    extended_ys = torch.cat([ys, zero_tensor], dim=1)

    for y in extended_ys:
        #eos_idx = (y==0).nonzero()[0].item()
        #print(y)
        eos_idx = torch.nonzero(y==0)[0].item()
        y[eos_idx] = eos
        y_out.append(y)

    y_out = pad_list(y_out)
    return y_in, y_out


from phonepiece.inventory import read_inventory
from transphone.g2p import read_g2p
from transphone.config import TransphoneConfig

class BaseTokenizer:

    def __init__(self, lang_id, g2p_model='latest', device=None):
        self.lang_id = lang_id
        self.inventory = read_inventory(lang_id)

        if g2p_model is None:
            self.g2p = None
        else:
            self.g2p = read_g2p(g2p_model, device)

        self.cache = {}

        if self.g2p is not None and self.g2p.cache_path is not None:
            lang_cache_path = self.g2p.cache_path / f"{lang_id}.txt"
            if lang_cache_path.exists():
                for line in open(lang_cache_path, 'r'):
                    fields = line.strip().split()
                    self.cache[fields[0]] = fields[1:]

        self.punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
        self.logger = TransphoneConfig.logger

    def tokenize(self, text: str):
        raise NotImplementedError

    def tokenize_words(self, text:str):
        text = text.translate(str.maketrans('', '', self.punctuation)).lower()

        return text.split()

    def convert_tokens_to_ids(self, lst):
        lst = list(filter(lambda s: s!='<SPACE>', lst))

        return self.inventory.phoneme.atoi(lst)

    def convert_ids_to_tokens(self, lst):
        return self.inventory.phoneme.itoa(lst)



from transphone.tokenizer import raw_epitran_dict
from transphone.lang.epitran_tokenizer import read_raw_epitran_tokenizer
from phonepiece.lexicon import read_lexicon
from phonepiece.distance import phonological_distance
import editdistance
from transphone.model.checkpoint_utils import *
import argparse

def eval_epitran(langs=None, exclude_langs=None):

    exp_dir = TransphoneConfig.data_path / 'decode' / 'epitran'
    exp_dir.mkdir(exist_ok=True, parents=True)

    log_w = open(exp_dir / f'result.md', 'w')

    log_w.write('| language | phoneme error rate | phonological distance |\n')
    log_w.write('|----------|--------------------|-----------------------|\n')

    if langs is not None and len(langs) != 0:
        target_langs = langs
    elif exclude_langs is not None and len(exclude_langs) != 0:
        target_langs = []
        for lang in  list(raw_epitran_dict.keys()):
            if lang not in exclude_langs:
                target_langs.append(lang)
    else:
        target_langs = list(raw_epitran_dict.keys())

    tot_cer = 0
    tot_fer = 0
    tot_csum = 0

    for lang in target_langs:
        cer = 0
        fer = 0
        csum = 0

        print("processing ", lang)

        epitran_id = raw_epitran_dict[lang]

        try:
            lexicon = read_lexicon(lang)
        except:
            print("could not read lexicon of ", lang)
            continue

        if len(lexicon) < 50:
            print("skipping ", lang)
            continue

        w = open(exp_dir / f'{lang}.txt', 'w')

        # at most 10000 to prevent overfitting
        word2phoneme_lst = list(lexicon.word2phoneme.items())

        try:
            epitran_tokenizer = read_raw_epitran_tokenizer(epitran_id, use_lexicon=False)
        except:
            print('failed to read epitran', lang)

        # last 10 for validation
        for grapheme_str, phonemes in word2phoneme_lst[-25:]:
            hyp_phonemes = epitran_tokenizer.tokenize(grapheme_str)
            w.write(f'lang: {lang}\n')
            w.write(f'inp : {grapheme_str}\n')
            w.write(f'ref : {" ".join(phonemes)}\n')
            w.write(f'hyp : {" ".join(hyp_phonemes)}\n')

            cur_cer = editdistance.distance(phonemes, hyp_phonemes)
            cur_fer = phonological_distance(phonemes, hyp_phonemes)
            cur_sum = len(phonemes)
            w.write(f'cer: {cur_cer/cur_sum}\n')
            w.write(f'fer: {cur_fer/cur_sum}\n\n')

            cer += cur_cer
            fer += cur_fer
            csum += cur_sum

        w.close()

        tot_cer += cer
        tot_fer += fer
        tot_csum += csum

        log_w.write(f'| {lang} | {cer / csum:.3f} | {fer / csum:.3f} |\n')

        print('lang   :', lang)
        print("val cer: ", cer / csum)
        print("val fer: ", fer / csum)

    print('all')
    print("val cer: ", tot_cer / tot_csum)
    print("val fer: ", tot_fer / tot_csum)
    log_w.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='eval epitran g2p')
    parser.add_argument('--lang', type=str, help='language')
    parser.add_argument('--exclude_lang', type=str, help='excluded_language')

    args = parser.parse_args()

    lang = set()
    exclude_lang = set()

    if args.lang is not None:
        lang.udpate(args.lang.split(','))
    if args.exclude_lang is not None:
        exclude_lang.update(args.exclude_lang.split(','))

    eval_epitran(lang, exclude_lang)






from transphone.utils import import_with_auto_install
from transphone.lang.jpn.kana2phoneme import Kana2Phoneme
from transphone.g2p import read_g2p
from phonepiece.inventory import read_inventory
from transphone.lang.jpn import jaconv
from transphone.lang.base_tokenizer import BaseTokenizer
from transphone.lang.jpn.normalizer import normalize_neologd, parse_jpn_number, parse_jpn_alphabet


def read_jpn_tokenizer(lang_id, g2p_model='latest', device=None, use_lexicon=True):
    return JPNTokenizer(lang_id, g2p_model, device)

class JPNTokenizer(BaseTokenizer):

    def __init__(self, lang_id, g2p_model='latest', device=None):

        super().__init__(lang_id, g2p_model, device)

        # import mecab and its dict
        MeCab = import_with_auto_install('MeCab', 'mecab-python3')
        import_with_auto_install('unidic_lite', 'unidic-lite')

        self.tagger = MeCab.Tagger()
        self.kana2phoneme = Kana2Phoneme()

    def tokenize_words(self, text:str):
        text = normalize_neologd(text)

        raw_words = self.tagger.parse(text).split('\n')

        result = []

        # exclude the last EOS word
        for word in raw_words[:-2]:
            raw = word.split('\t')[0]
            result.append(raw)

        return result

    def tokenize(self, text, use_g2p=True, use_space=False, verbose=False):

        text = normalize_neologd(text)

        raw_words = self.tagger.parse(text).split('\n')

        result = []

        # exclude the last EOS word
        for word in raw_words[:-2]:

            kana = word.split('\t')[1]
            raw = word.split('\t')[0]

            if str.isdigit(raw):
                hankaku_num = jaconv.z2h(kana)
                kana = jaconv.hira2kata(parse_jpn_number(hankaku_num))

            # interestingly, isalpha also return true for kana
            if str.isalpha(raw) and str.isascii(raw):
                kana = parse_jpn_alphabet(raw)

            res = self.kana2phoneme.convert(kana)

            if res == ['*'] and self.kana2phoneme.validate(raw):
                res = self.kana2phoneme.convert(raw)

            if verbose:
                print(kana, res)

            if res != ['*']:
                result.extend(res)
            elif use_g2p:
                if raw in self.cache:
                    result.extend(self.cache[raw])
                else:
                    phonemes = self.g2p.inference(raw)
                    remapped_phonemes = self.inventory.remap(phonemes)
                    if verbose:
                        print(f"g2p {raw} ->  {remapped_phonemes}")
                    self.cache[raw] = remapped_phonemes
                    result.extend(remapped_phonemes)

            if use_space:
                result.append('<SPACE>')

        return result


from transphone.utils import import_with_auto_install
from transphone.lang.base_tokenizer import BaseTokenizer
from phonepiece.arpa import ArpaConverter
from transphone.lang.eng.normalizer import ENGNormalizer

def read_eng_tokenizer(lang_id='eng', g2p_model='latest', device=None, use_lexicon=True):
    return ENGTokenizer(lang_id, g2p_model, device)


class ENGTokenizer(BaseTokenizer):

    def __init__(self, lang_id='eng', g2p_model='latest', device=None):

        super().__init__(lang_id, g2p_model, device)

        # import jieba and pypinyin for segmentation
        cmudict_module = import_with_auto_install('cmudict', 'cmudict')
        self.cmudict = cmudict_module.dict()
        self.converter = ArpaConverter()
        self.normalizer = ENGNormalizer()

    def tokenize(self, text, use_g2p=True, use_space=False, verbose=False):

        norm_text = self.normalizer(text)

        log = f"normalization: {text} -> {norm_text}"
        self.logger.info(log)
        if verbose:
            print(log)

        ipa_lst = []
        text = norm_text

        for word in text.split():
            if len(word) >= 1:
                if word in self.cache:
                    ipas = self.cache[word]
                    ipa_lst.extend(ipas)

                elif word in self.cmudict:
                    arpa = self.cmudict[word][0]
                    ipas = self.converter.convert(arpa)
                    ipa_lst.extend(ipas)

                    log = f"CMUdict: {word} -> {arpa} -> {ipas}"
                    self.logger.info(f"CMUdict: {word} -> {arpa} -> {ipas}")
                    if verbose:
                        print(log)

                    self.cache[word] = ipas

                elif use_g2p:
                    phonemes = self.g2p.inference(word)
                    remapped_phonemes = self.inventory.remap(phonemes)

                    log = f"g2p {word} ->  {remapped_phonemes}"
                    self.logger.info(log)
                    if verbose:
                        print(log)
                    self.cache[word] = remapped_phonemes
                    ipa_lst.extend(remapped_phonemes)
                if use_space:
                    ipa_lst.append('<SPACE>')

        return ipa_lst


from transphone.model.loader import read_loader
from transphone.model.dataset import read_multilingual_dataset
from transphone.model.transformer import TransformerG2P
from transphone.model.utils import read_model_config
import torch.nn as nn
from torch.utils.tensorboard import SummaryWriter
import editdistance
import tqdm
from transphone.model.checkpoint_utils import *
import argparse


def train(exp, checkpoint):

    config = read_model_config(exp)

    model_dir = TransphoneConfig.data_path / "model" / exp
    model_dir.mkdir(parents=True, exist_ok=True)

    train_dataset, test_dataset = read_multilingual_dataset()
    SRC_VOCAB_SIZE = len(train_dataset.grapheme_vocab)+1
    TGT_VOCAB_SIZE = len(train_dataset.phoneme_vocab)+1

    print("src vocab size ", SRC_VOCAB_SIZE)
    print("tgt vocab size ", TGT_VOCAB_SIZE)

    train_dataset.grapheme_vocab.write(model_dir / 'grapheme.vocab')
    train_dataset.phoneme_vocab.write(model_dir / 'phoneme.vocab')

    EMB_SIZE = config.embed_size
    NHEAD = config.num_head
    FFN_HID_DIM = config.hidden_size
    NUM_ENCODER_LAYERS = config.num_encoder
    NUM_DECODER_LAYERS = config.num_decoder
    torch.manual_seed(0)

    model = TransformerG2P(NUM_ENCODER_LAYERS, NUM_DECODER_LAYERS, EMB_SIZE,
                            NHEAD, SRC_VOCAB_SIZE, TGT_VOCAB_SIZE, FFN_HID_DIM).cuda()


    for p in model.parameters():
        if p.dim() > 1:
            nn.init.xavier_uniform_(p)

    if checkpoint is not None:
        torch_load(model, checkpoint)

    opt = torch.optim.Adam(model.parameters(), lr=0.0001, betas=(0.9, 0.98), eps=1e-9)

    print('training size: ', len(train_dataset))
    print("testing size: ", len(test_dataset))

    train_loader = read_loader(train_dataset, batch_size=64)
    test_loader = read_loader(test_dataset, batch_size=1)

    epoch = 400
    writer = SummaryWriter()

    iteration = 0

    best_cer = 2.0

    for i in range(epoch):

        it = iter(train_loader)

        loss_sum = 0
        it_cnt = len(it)

        for batch in tqdm.tqdm(it):

            x,y = batch
            #print(x)
            #print(y)
            #print(torch.max(x).item(), torch.max(y).item())

            x = x.cuda()
            y = y.cuda()

            opt.zero_grad()
            loss = model.train_step(x, y)
            loss.backward()
            opt.step()

            loss_val = loss.item()
            iteration += 1
            #print(iteration, ' ', loss_val)
            writer.add_scalar('Loss/Train', loss_val, iteration)
            loss_sum += loss

        print("epoch ", i, " loss ", loss_sum/it_cnt)

        test_it = iter(test_loader)

        cer = 0
        csum = 0
        wer = 0
        wsum = 0

        ploted =False

        if i % 5 != 0:
            continue

        for x,y in test_it:
            x = x.cuda()
            y = y.cuda()
            predicted = model.inference(x)
            y = y.squeeze(0).tolist()

            dist = editdistance.eval(predicted, y)
            cer += dist
            csum += len(y)

            if dist != 0:
                wer += 1
            wsum += 1

        writer.add_scalar('CER/Test', cer/csum, i)
        writer.add_scalar('WER/Test', wer/wsum, i)

        print("val cer: ", cer/csum)
        print("val wer: ", wer/wsum)

        cer = cer/csum

        if cer <= best_cer:
            best_cer = cer
            model_path = model_dir / f"model_{best_cer:0.6f}.pt"
            torch_save(model, model_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='train g2p')
    parser.add_argument('--exp', type=str, help='exp name')
    parser.add_argument('--checkpoint', type=str, help='checkpoint')

    args = parser.parse_args()

    exp = args.exp
    checkpoint=args.checkpoint

    train(exp, checkpoint)



